"""Constants for the Layer Values Monitor."""

import os
from collections import deque
from datetime import UTC, datetime

DENOM = "loya"

TABLE = deque(maxlen=60)
CSV_FILE_PATTERN = "table_{timestamp}.csv"
CURRENT_CSV_FILE = f"table_{int(datetime.now(UTC).timestamp())}.csv"
LOGS_DIR = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "logs")



import asyncio
import json
import os
import tempfile
import time
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

from layer_values_monitor.config_watcher import ConfigWatcher

import pytest
from dotenv import load_dotenv


@pytest.fixture
def mock_websocket():
    """Fixture for a mock WebSocket connection."""
    mock_ws = AsyncMock()
    mock_ws.send = AsyncMock()
    mock_ws.recv = AsyncMock()
    return mock_ws


@pytest.fixture
def mock_websockets_connect(mock_websocket):
    mock_connect = MagicMock()
    mock_connect.return_value.__aenter__.return_value = mock_websocket
    patcher = patch("websockets.connect", mock_connect)

    mock = patcher.start()
    yield mock
    patcher.stop()


@pytest.fixture
def event_queue():
    return asyncio.Queue()


@pytest.fixture
def disputes_queue():
    return asyncio.Queue()


@pytest.fixture
def test_report_messages():
    """Fixture that loads test report messages from a JSON file."""
    base_dir = os.path.dirname(os.path.abspath(__file__))
    json_path = os.path.join(base_dir, "test_messages.json")

    with open(json_path) as f:
        return [json.dumps(msg) for msg in json.load(f)]


@pytest.fixture(scope="session", autouse=True)
def load_env():
    load_dotenv(".env", override=True)


@pytest.fixture
def config_file():
    """Create a temporary config file for testing."""
    # Create a temporary file for testing
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=".toml")
    temp_file_path = Path(temp_file.name)

    # Write initial config to the file
    with open(temp_file_path, "w") as f:
        f.write("""
            [Feed]
            test = "initial_value"
            
            [Settings]
            interval = 5
        """)

    # Allow filesystem to register the file
    time.sleep(0.1)

    yield temp_file_path

    # Clean up after test
    os.unlink(temp_file_path)


@pytest.fixture
def config_watcher(config_file):
    """Create a ConfigWatcher instance for testing."""
    return ConfigWatcher(config_file)



"""Layer values monitor."""

import argparse
import asyncio
import logging
import os
from functools import wraps
from pathlib import Path
from typing import Any

from layer_values_monitor.config_watcher import ConfigWatcher, watch_config
from layer_values_monitor.dispute import process_disputes
from layer_values_monitor.logger import logger
from layer_values_monitor.monitor import (
    listen_to_new_report_events,
    new_reports_queue_handler,
    raw_data_queue_handler,
)
from layer_values_monitor.threshold_config import ThresholdConfig

from dotenv import load_dotenv
from telliot_core.apps.telliot_config import TelliotConfig

for logger_name in logging.root.manager.loggerDict:
    if logger_name.startswith("telliot_feeds"):
        logging.getLogger(logger_name).setLevel(logging.CRITICAL)


def async_run(f: Any) -> Any:
    """Wrap an async function to be called synchronously and handle keyboard interrupts."""

    @wraps(f)
    def wrapper(*args, **kwargs):  # noqa: ANN202
        try:
            return asyncio.run(f(*args, **kwargs))
        except KeyboardInterrupt:
            print("Exiting...")

    return wrapper


@async_run
async def start() -> None:
    """Start layer values monitor."""
    loaded = load_dotenv(override=True)
    if not loaded:
        raise ValueError("Failed to load environment variables")
    uri = os.getenv("URI")
    if uri is None:
        raise ValueError("URI not found in environment variables")
    chain_id = os.getenv("CHAIN_ID")
    if chain_id is None:
        raise ValueError("CHAIN_ID not found in environment variables")
    parser = argparse.ArgumentParser(description="Start values monitor")
    parser.add_argument("binary_path", type=str, help="Path to the Layer binary executable")
    parser.add_argument("key_name", type=str, help="Name of the key to use for transactions")
    parser.add_argument("keyring_backend", type=str, help="Keyring backend")
    parser.add_argument("keyring_dir", type=str, help="Keyring directory")
    parser.add_argument("--payfrom-bond", action="store_true", help="Pay dispute fee from bond")
    parser.add_argument("--use-custom-config", action="store_true", help="Use custom config.toml")
    # percentage
    parser.add_argument("--global-percentage-alert-threshold", type=float, help="Global percent threshold")
    parser.add_argument(
        "--global-percentage-warning-threshold", default=0.0, type=float, help="Global percentage for a warning dispute cat"
    )
    parser.add_argument(
        "--global-percentage-minor-threshold", default=0.0, type=float, help="Global percentage for a minor dispute cat"
    )
    parser.add_argument(
        "--global-percentage-major-threshold", default=0.0, type=float, help="Global percentage for a major dispute cat"
    )
    # range
    parser.add_argument("--global-range-alert-threshold", default=0.0, type=float, help="Global range threshold")
    parser.add_argument(
        "--global-range-warning-threshold", default=0.0, type=float, help="Global range for a warning dispute cat"
    )
    parser.add_argument(
        "--global-range-minor-threshold", default=0.0, type=float, help="Global range for a minor dispute cat"
    )
    parser.add_argument(
        "--global-range-major-threshold", default=0.0, type=float, help="Global range for a major dispute cat"
    )
    # equality
    parser.add_argument("--global-equality-alert-threshold", default=False, type=bool, help="Global equality threshold")
    parser.add_argument(
        "--global-equality-warning-threshold", default=False, type=bool, help="Global equality for a warning dispute cat"
    )
    parser.add_argument(
        "--global-equality-minor-threshold", default=False, type=bool, help="Global equality for a minor dispute cat"
    )
    parser.add_argument(
        "--global-equality-major-threshold", default=False, type=bool, help="Global equality for a major dispute cat"
    )
    args = parser.parse_args()

    threshold_config = ThresholdConfig.from_args(args)

    # Initialize config watcher
    config_path = Path(__file__).resolve().parents[2] / "config.toml"
    config_watcher = ConfigWatcher(config_path)

    raw_data_queue = asyncio.Queue()
    new_reports_queue = asyncio.Queue()
    disputes_queue = asyncio.Queue()
    cfg = TelliotConfig()
    cfg.main.chain_id = 1

    # TODO: validate user options to check if they conflict
    try:
        await asyncio.gather(
            listen_to_new_report_events(uri, raw_data_queue, logger),
            raw_data_queue_handler(
                raw_data_queue,
                new_reports_queue,
                logger=logger,
            ),
            new_reports_queue_handler(new_reports_queue, disputes_queue, config_watcher, logger, threshold_config),
            process_disputes(
                disputes_q=disputes_queue,
                binary_path=args.binary_path,
                key_name=args.key_name,
                kb=args.keyring_backend,
                kdir=args.keyring_dir,
                rpc=f"http://{uri}",
                chain_id=chain_id,
                payfrom_bond=args.payfrom_bond,
                logger=logger,
            ),
            watch_config(config_watcher),
        )
    except asyncio.CancelledError:
        print("shutting down running tasks")
        raise


if __name__ == "__main__":
    start()



"""Module to watch and manage a live configuration."""

import asyncio
import time
import tomllib
from pathlib import Path
from typing import Any

from layer_values_monitor.logger import logger


class ConfigWatcher:
    """Class to watch and manage a live configuration."""

    def __init__(self, config_path: Path) -> None:
        """Initialize the configuration watcher."""
        self.config_path = config_path
        self.config = {}
        self.last_modified_time = 0
        self.reload_config()

    def reload_config(self) -> bool:
        """Reload config if modified, return True if reloaded."""
        current_mtime = self.config_path.stat().st_mtime
        if current_mtime > self.last_modified_time:
            with open(self.config_path, "rb") as f:
                data = tomllib.load(f)
                data = {k.lower(): v for k, v in data.items()}
                for key, value in data.items():
                    if isinstance(value, dict):
                        data[key.lower()] = {k.lower(): v for k, v in value.items()}
            self.config = data
            self.last_modified_time = current_mtime
            logger.info(f"Configuration reloaded at {time.strftime('%H:%M:%S')}")
            return True
        return False

    def get_config(self) -> dict[str, Any]:
        """Get the current configuration."""
        return self.config


async def watch_config(config_watcher: ConfigWatcher, check_interval: float = 5.0) -> None:
    """Watch the config file for changes and reload when modified."""
    while True:
        config_watcher.reload_config()
        await asyncio.sleep(check_interval)



import asyncio
import time

from layer_values_monitor.main import ConfigWatcher, watch_config

import pytest


def test_init(config_file, config_watcher):
    """Test the initialization of ConfigWatcher."""
    assert isinstance(config_watcher, ConfigWatcher)
    assert config_watcher.config_path == config_file
    assert config_watcher.last_modified_time > 0


def test_get_config(config_watcher):
    """Test that config is correctly loaded and accessible."""
    config = config_watcher.get_config()
    assert "feed" in config
    assert config["feed"]["test"] == "initial_value"
    assert "settings" in config
    assert config["settings"]["interval"] == 5


def test_reload_config_when_changed(config_file, config_watcher):
    """Test that config is reloaded when file changes."""
    # Record initial timestamp
    initial_timestamp = config_watcher.last_modified_time

    # Wait a bit to ensure file timestamp will be different
    time.sleep(0.1)

    # Modify config file
    with open(config_file, "w") as f:
        f.write("""
            [Feed]
            test = "updated_value"
            new_key = "new_value"
            
            [Settings]
            interval = 10
        """)

    # Reload config
    reloaded = config_watcher.reload_config()

    # Check that reload happened
    assert reloaded is True
    assert config_watcher.last_modified_time > initial_timestamp

    # Check that config was updated
    config = config_watcher.get_config()
    assert config["feed"]["test"] == "updated_value"
    assert config["feed"]["new_key"] == "new_value"
    assert config["settings"]["interval"] == 10


def test_reload_config_when_unchanged(config_watcher):
    """Test that reload returns False when file hasn't changed."""
    # Call reload_config once to set initial state
    config_watcher.reload_config()

    # Call again without changing the file
    reloaded = config_watcher.reload_config()

    # Should return False as file hasn't changed
    assert reloaded is False


@pytest.mark.asyncio
async def test_watch_config(config_file, config_watcher):
    """Test the watch_config async function."""
    # Set a very short check interval for testing
    check_interval = 0.1

    # Create a task for watch_config with a short timeout
    task = asyncio.create_task(watch_config(config_watcher, check_interval))

    # Wait a bit to let watch_config run
    await asyncio.sleep(check_interval * 2)

    # Record the current state
    initial_timestamp = config_watcher.last_modified_time
    initial_config = config_watcher.get_config()

    # Modify config file
    with open(config_file, "w") as f:
        f.write("""
            [Feed]
            test = "auto_updated"
            
            [Settings]
            interval = 15
        """)

    # Wait for watch_config to detect the change
    await asyncio.sleep(check_interval * 2)

    # Cancel the task
    task.cancel()

    try:
        await task
    except asyncio.CancelledError:
        pass

    # Verify that config was automatically updated
    assert config_watcher.last_modified_time > initial_timestamp
    assert config_watcher.get_config() != initial_config
    assert config_watcher.get_config()["feed"]["test"] == "auto_updated"
    assert config_watcher.get_config()["settings"]["interval"] == 15



"""LVM helper functions."""

import logging
import os
from datetime import UTC, datetime

from layer_values_monitor.constants import CSV_FILE_PATTERN, CURRENT_CSV_FILE, LOGS_DIR, TABLE
from layer_values_monitor.custom_types import GlobalMetric, Metrics
from layer_values_monitor.threshold_config import ThresholdConfig

import pandas as pd
from dotenv import load_dotenv
from pandas import DataFrame

# Load environment variables
load_dotenv()
MAX_TABLE_ROWS = int(os.getenv("MAX_TABLE_ROWS", "100000"))

def get_current_csv_path() -> str:
    """Get the full path to the current CSV file."""
    return os.path.join(LOGS_DIR, CURRENT_CSV_FILE)

def should_create_new_file() -> bool:
    """Check if we should create a new file based on row count."""
    current_file = get_current_csv_path()
    if not os.path.exists(current_file):
        return True
    
    try:
        # Read the CSV file and count rows (excluding header)
        df = pd.read_csv(current_file)
        return len(df) >= MAX_TABLE_ROWS
    except Exception as e:
        logging.error(f"Error checking file size: {e}")
        return False

def create_new_csv_file() -> str:
    """Create a new CSV file with current timestamp and return its path."""
    timestamp = int(datetime.now(UTC).timestamp())
    new_filename = CSV_FILE_PATTERN.format(timestamp=timestamp)
    new_filepath = os.path.join(LOGS_DIR, new_filename)
    
    # Update the current CSV file constant
    global CURRENT_CSV_FILE
    CURRENT_CSV_FILE = new_filename
    
    return new_filepath

def get_metric(
    query_type: str,
    logger: logging,
    threshold_config: ThresholdConfig,
) -> Metrics | None:
    """Return Metrics/Thresholds given a query type."""
    try:
        metric = GlobalMetric[query_type.upper()].value
    except KeyError:
        logger.error(f"query type {query_type} not found in global metrics")
        print(f"ERROR: query type {query_type} not found in global metrics")
        return None
    if metric == "percentage":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.percentage_alert,
            warning_threshold=threshold_config.percentage_warning,
            minor_threshold=threshold_config.percentage_minor,
            major_threshold=threshold_config.percentage_major,
        )
    elif metric == "range":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.range_alert,
            warning_threshold=threshold_config.range_warning,
            minor_threshold=threshold_config.range_minor,
            major_threshold=threshold_config.range_major,
        )
    elif metric == "equality":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.equality_alert,
            warning_threshold=threshold_config.equality_warning,
            minor_threshold=threshold_config.equality_minor,
            major_threshold=threshold_config.equality_major,
        )


def remove_0x_prefix(s: str) -> str:
    """Remove 0x prefix if there from hex string."""
    if s.startswith("0x"):
        return s[2:]
    return s


def add_to_table(entry: dict[str, str]) -> None:
    """Add entry to table and print it."""
    TABLE.append(entry)
    os.system("clear")
    df = DataFrame(TABLE).sort_values(by="TIMESTAMP")
    print(df.to_string(index=False, justify="center"))

    # Check if we need to create a new file
    if should_create_new_file():
        csv_file = create_new_csv_file()
        # Create new file with header
        DataFrame([entry]).to_csv(csv_file, mode="w", header=True, index=False)
    else:
        csv_file = get_current_csv_path()
        # Append to existing file
        DataFrame([entry]).to_csv(csv_file, mode="a", header=False, index=False)



import asyncio
import logging
from unittest.mock import MagicMock

from layer_values_monitor.custom_types import NewReport
from layer_values_monitor.monitor import raw_data_queue_handler

import pytest


@pytest.fixture
def mock_logger():
    """Create a mock logger for testing."""
    logger = MagicMock()
    logger.info = MagicMock()
    logger.warning = MagicMock()
    logger.error = MagicMock()
    logger.debug = MagicMock()
    return logger


@pytest.mark.asyncio
async def test_raw_data_queue_handler_basic_flow(mock_logger):
    """Test the basic flow of the raw_data_queue_handler function."""
    raw_data_q = asyncio.Queue()
    new_reports_q = asyncio.Queue()

    # Create sample raw data with different block heights
    sample_data_1 = {
        "result": {
            "events": {
                "tx.height": ["100"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["query_id_1"],
                "new_report.value": ["0x123"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle1"],
                "new_report.reporter_power": ["1000"],
                "new_report.reporter": ["reporter1"],
                "new_report.timestamp": ["1625097600000"],
                "new_report.meta_id": ["meta1"],
                "tx.hash": ["hash1"],
            }
        }
    }

    # Same block height, different query_id
    sample_data_2 = {
        "result": {
            "events": {
                "tx.height": ["100"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["query_id_2"],
                "new_report.value": ["0x456"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle1"],
                "new_report.reporter_power": ["2000"],
                "new_report.reporter": ["reporter2"],
                "new_report.timestamp": ["1625097600001"],
                "new_report.meta_id": ["meta2"],
                "tx.hash": ["hash2"],
            }
        }
    }

    # Higher block height
    sample_data_3 = {
        "result": {
            "events": {
                "tx.height": ["101"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["query_id_1"],
                "new_report.value": ["0x789"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle2"],
                "new_report.reporter_power": ["3000"],
                "new_report.reporter": ["reporter3"],
                "new_report.timestamp": ["1625097600002"],
                "new_report.meta_id": ["meta3"],
                "tx.hash": ["hash3"],
            }
        }
    }

    sample_data_4 = {
        "result": {
            "events": {
                "tx.height": ["102"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["query_id_3"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.value": ["0x789"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle2"],
                "new_report.reporter_power": ["3000"],
                "new_report.reporter": ["reporter3"],
                "new_report.timestamp": ["1625097600002"],
                "new_report.meta_id": ["meta3"],
                "tx.hash": ["hash3"],
            }
        }
    }

    sample_data_5 = {"result": None}
    sample_data_6 = {"result": {"no_events": True}}

    for data in [sample_data_1, sample_data_2, sample_data_3, sample_data_4, sample_data_5, sample_data_6]:
        await raw_data_q.put(data)

    await raw_data_queue_handler(raw_data_q, new_reports_q, mock_logger, max_iterations=6)

    assert new_reports_q.qsize() == 2

    # Get first collection (block height 100)
    first_collection = await new_reports_q.get()

    # The first collection should have 2 query_ids
    assert len(first_collection) == 2
    assert "query_id_1" in first_collection
    assert "query_id_2" in first_collection
    assert isinstance(first_collection["query_id_1"][0], NewReport)
    assert first_collection["query_id_1"][0].query_id == "query_id_1"
    assert first_collection["query_id_1"][0].value == "0x123"
    assert first_collection["query_id_1"][0].reporter == "reporter1"

    # Check the second report in the first collection
    assert isinstance(first_collection["query_id_2"][0], NewReport)
    assert first_collection["query_id_2"][0].query_id == "query_id_2"
    assert first_collection["query_id_2"][0].value == "0x456"
    assert first_collection["query_id_2"][0].reporter == "reporter2"

    # Get second collection (block height 101)
    second_collection = await new_reports_q.get()

    assert len(second_collection) == 1
    assert "query_id_1" in second_collection

    # Check the report in the second collection
    assert isinstance(second_collection["query_id_1"][0], NewReport)
    assert second_collection["query_id_1"][0].query_id == "query_id_1"
    assert second_collection["query_id_1"][0].value == "0x789"
    assert second_collection["query_id_1"][0].reporter == "reporter3"


@pytest.mark.asyncio
async def test_raw_data_queue_handler_empty_queue(mock_logger):
    """Test handling of an empty queue."""
    raw_data_q = asyncio.Queue()
    new_reports_q = asyncio.Queue()

    # Create and run the task with a short timeout
    task = asyncio.create_task(raw_data_queue_handler(raw_data_q, new_reports_q, mock_logger, max_iterations=1))

    await raw_data_q.put({})

    await asyncio.sleep(0.1)

    assert new_reports_q.qsize() == 0

    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


@pytest.mark.asyncio
async def test_raw_data_queue_handler_sequential_same_height(mock_logger):
    """Test handling multiple reports with the same block height."""
    raw_data_q = asyncio.Queue()
    new_reports_q = asyncio.Queue()

    task = asyncio.create_task(raw_data_queue_handler(raw_data_q, new_reports_q, mock_logger, max_iterations=5))

    # Wait for the task to start
    await asyncio.sleep(0.1)

    # Add a single report with block height 200
    report1 = {
        "result": {
            "events": {
                "tx.height": ["200"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["same_query_id"],
                "new_report.value": ["0x111"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle1"],
                "new_report.reporter_power": ["1000"],
                "new_report.reporter": ["reporter1"],
                "new_report.timestamp": ["1625097600000"],
                "new_report.meta_id": ["meta"],
                "tx.hash": ["hash1"],
            }
        }
    }
    await raw_data_q.put(report1)

    # Add a second report with the same block height
    report2 = {
        "result": {
            "events": {
                "tx.height": ["200"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["same_query_id"],
                "new_report.value": ["0x222"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle1"],
                "new_report.reporter_power": ["2000"],
                "new_report.reporter": ["reporter2"],
                "new_report.timestamp": ["1625097600000"],
                "new_report.meta_id": ["meta"],
                "tx.hash": ["hash2"],
            }
        }
    }
    await raw_data_q.put(report2)

    # Now add a report with a higher block height to trigger collection clearing
    final_report = {
        "result": {
            "events": {
                "tx.height": ["201"],
                "new_report.query_type": ["SpotPrice"],
                "new_report.query_data": ["0x..."],
                "new_report.query_id": ["other_query_id"],
                "new_report.value": ["0xfinal"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["cycle1"],
                "new_report.reporter_power": ["3000"],
                "new_report.reporter": ["reporter3"],
                "new_report.timestamp": ["1625097600000"],
                "new_report.meta_id": ["meta"],
                "tx.hash": ["hash3"],
            }
        }
    }
    await raw_data_q.put(final_report)

    # Wait for the collection to be processed
    await asyncio.sleep(0.5)

    # Now there should be a collection in new_reports_q
    assert new_reports_q.qsize() >= 1

    # Get the collection
    collection = await new_reports_q.get()

    assert "same_query_id" in collection

    # It should have exactly 2 reports
    assert len(collection["same_query_id"]) == 2

    report_values = [report.value for report in collection["same_query_id"]]
    assert "0x111" in report_values
    assert "0x222" in report_values

    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


@pytest.mark.asyncio
async def test_raw_data_queue_handler():
    """Test the raw_data_queue_handler function processes reports correctly."""

    raw_data_q = asyncio.Queue()
    new_reports_q = asyncio.Queue()
    logger = MagicMock(spec=logging.Logger)

    raw_data_height_5 = {
        "result": {
            "events": {
                "tx.height": ["5"],
                "new_report.query_type": ["price"],
                "new_report.query_data": ["BTC/USD"],
                "new_report.query_id": ["query1"],
                "new_report.value": ["50000"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["12"],
                "new_report.reporter_power": ["100"],
                "new_report.reporter": ["reporter1"],
                "new_report.timestamp": ["1620000000"],
                "new_report.meta_id": ["meta1"],
                "tx.hash": ["hash1"],
            }
        }
    }

    raw_data_height_5_query2 = {
        "result": {
            "events": {
                "tx.height": ["5"],
                "new_report.query_type": ["price"],
                "new_report.query_data": ["ETH/USD"],
                "new_report.query_id": ["query2"],
                "new_report.value": ["3000"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["12"],
                "new_report.reporter_power": ["100"],
                "new_report.reporter": ["reporter1"],
                "new_report.timestamp": ["1620000000"],
                "new_report.meta_id": ["meta2"],
                "tx.hash": ["hash2"],
            }
        }
    }

    raw_data_height_6 = {
        "result": {
            "events": {
                "tx.height": ["6"],
                "new_report.query_type": ["price"],
                "new_report.query_data": ["BTC/USD"],
                "new_report.query_id": ["query1"],
                "new_report.value": ["52000"],
                "new_report.aggregate_method": ["median"],
                "new_report.cyclelist": ["12"],
                "new_report.reporter_power": ["300"],
                "new_report.reporter": ["reporter3"],
                "new_report.timestamp": ["1620000020"],
                "new_report.meta_id": ["meta4"],
                "tx.hash": ["hash4"],
            }
        }
    }

    await raw_data_q.put(raw_data_height_5)
    await raw_data_q.put(raw_data_height_5_query2)
    await raw_data_q.put(raw_data_height_6)

    await raw_data_queue_handler(raw_data_q, new_reports_q, logger, max_iterations=3)

    assert not new_reports_q.empty(), "Queue should have items"
    height_5_reports = await new_reports_q.get()

    assert len(height_5_reports) == 2, "Should have reports for 2 query IDs"
    assert "query1" in height_5_reports, "Should have reports for query1"
    assert "query2" in height_5_reports, "Should have reports for query2"

    # Verify query1 for height 5
    query1_reports = height_5_reports["query1"]
    assert len(query1_reports) == 1, "query1 should have 1 report"
    assert query1_reports[0].value == "50000", "Report should have value 50000"

    # Verify query2 for height 5
    query2_reports = height_5_reports["query2"]
    assert len(query2_reports) == 1, "query2 should have 1 report"
    assert query2_reports[0].value == "3000", "Report should have value 3000"



"""Propose dispute."""

import asyncio
import json
import logging
import subprocess
import time

from layer_values_monitor.custom_types import DisputeCategory, Msg
from layer_values_monitor.logger import logger


def propose_msg(
    binary_path: str,
    reporter: str,
    query_id: str,
    meta_id: str,
    dispute_category: DisputeCategory,
    fee: str,
    key_name: str,
    chain_id: str,
    rpc: str,
    kb: str,
    kdir: str,
    payfrom_bond: str,
) -> str | None:
    """Execute propose-dispute message using layer's binary."""
    cmd = [
        binary_path,
        "tx",
        "dispute",
        "propose-dispute",
        reporter,
        meta_id,
        query_id,
        dispute_category,
        fee,
        payfrom_bond,
        "--from",
        key_name,
        "--chain-id",
        chain_id,
        "--node",
        rpc,
        "--keyring-backend",
        kb,
        "--keyring-dir",
        kdir,
        "--gas-prices",
        "1loya",
        "--gas",
        "auto",
        "--gas-adjustment",
        "1.3",
        "-y",
        "--output",
        "json",
    ]

    time.sleep(5)
    try:
        result = subprocess.run(cmd, capture_output=True)
        if result.returncode != 0:
            logger.error("Error calling dispute transaction in cli:", result.stderr)
            return None
        signed_tx = json.loads(result.stdout)
        code = signed_tx["code"]
        if code != 0:
            print(signed_tx)
            logger.error(f"failed to execute dispute msg: {signed_tx['raw_log']}")
            return None
        logger.info(f"dispute msg executed successfully: {signed_tx['txhash']}")
        return signed_tx["txhash"]
    except Exception as e:
        logger.error(f"failed to execute dispute msg: {e.__str__()}")
        return None


def determine_dispute_category(
    diff: float,
    # should be already manually sorted
    category_thresholds: dict[DisputeCategory, float],
) -> DisputeCategory | None:
    """Determine dispute category based on difference value and category thresholds."""
    # Return the most severe category whose threshold is met
    for category, threshold in category_thresholds.items():
        if threshold == 0:
            continue
        if diff >= threshold:
            return category

    return None


def determine_dispute_fee(
    category: DisputeCategory,
    reporter_power: int,
) -> int:
    """Calculate dispute fee based on category and reporter power."""
    if category == "warning":
        percentage = 0.01
    elif category == "minor":
        percentage = 0.05
    else:
        percentage = 1

    return int((reporter_power * 1_000_000) * percentage)


async def process_disputes(
    disputes_q: asyncio.Queue,
    binary_path: str,
    key_name: str,
    chain_id: str,
    rpc: str,
    kb: str,
    kdir: str,
    payfrom_bond: bool,
    logger: logging,
) -> None:
    """Process dispute messages from queue and submit them to the blockchain."""
    while True:
        dispute: Msg = await disputes_q.get()
        disputes_q.task_done()
        if dispute is None:
            continue
        time.sleep(2)
        logger.info(
            f"sending a dispute msg to layer:\n \
                    Reporter: {dispute.reporter}\n \
                    Query ID: {dispute.query_id} \
                    "
        )
        _ = propose_msg(
            binary_path=binary_path,
            key_name=key_name,
            chain_id=chain_id,
            rpc=rpc,
            kb=kb,
            kdir=kdir,
            reporter=dispute.reporter,
            query_id=dispute.query_id,
            meta_id=dispute.meta_id,
            dispute_category=dispute.category,
            fee=dispute.fee,
            payfrom_bond=str(payfrom_bond),
        )


if __name__ == "__main__":
    import shutil

    tx_hash = propose_msg(
        binary_path=shutil.which("layerd"),
        reporter="tellor1atxszkp3ar3gshqklhafd6rtumndz73zwfe0dx",
        meta_id="1",
        query_id="0xasdk",
        dispute_category="warning",
        fee="1000000loya",
        key_name="alice",
        chain_id="layer-1",
        rpc="http://localhost:26657",
        kb="test",
        kdir="~/.layer",
        payfrom_bond="True",
    )
    print(tx_hash)



import json
from unittest import mock

from layer_values_monitor.dispute import (
    determine_dispute_category,
    determine_dispute_fee,
    propose_msg,
)


def test_returns_correct_category_based_on_thresholds():
    category_thresholds = {
        "major": 0.5,
        "minor": 0.2,
        "warning": 0.1,
    }

    # Test when diff meets threshold for major
    assert determine_dispute_category(0.6, category_thresholds) == "major"

    # Test when diff meets threshold for minor but not major
    assert determine_dispute_category(0.3, category_thresholds) == "minor"

    # Test when diff meets threshold for warning but not minor
    assert determine_dispute_category(0.15, category_thresholds) == "warning"

    # Test when diff doesn't meet any threshold
    assert determine_dispute_category(0.05, category_thresholds) is None


def test_returns_none_when_all_thresholds_are_zero():
    zero_thresholds = {
        "major": 0,
        "minor": 0,
        "warning": 0,
    }
    assert determine_dispute_category(0.1, zero_thresholds) is None


def test_calculates_correct_fee_for_each_category():
    # Test fee calculation for warning category (1%)
    assert determine_dispute_fee("warning", 1000) == 10_000_000  # 1% of 1000 * 1,000,000

    # Test fee calculation for minor category (5%)
    assert determine_dispute_fee("minor", 1000) == 50_000_000  # 5% of 1000 * 1,000,000

    # Test fee calculation for major category (100%)
    assert determine_dispute_fee("major", 1000) == 1_000_000_000  # 100% of 1000 * 1,000,000


def test_returns_zero_when_reporter_power_is_zero():
    assert determine_dispute_fee("warning", 0) == 0
    assert determine_dispute_fee("minor", 0) == 0
    assert determine_dispute_fee("major", 0) == 0


@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_successful_execution(mock_subprocess_run):
    # Mock successful execution
    mock_process = mock.MagicMock()
    mock_process.returncode = 0
    mock_process.stdout = json.dumps({"code": 0, "txhash": "ABC123"}).encode()
    mock_subprocess_run.return_value = mock_process

    result = propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    # Assert the result is the txhash
    assert result == "ABC123"

    # Assert subprocess.run was called with the correct arguments
    mock_subprocess_run.assert_called_once()
    args, _ = mock_subprocess_run.call_args
    cmd = args[0]

    assert cmd[0] == "layerd"
    assert "tx" in cmd
    assert "dispute" in cmd
    assert "propose-dispute" in cmd
    assert "reporter_address" in cmd
    assert "meta_id" in cmd
    assert "query_id" in cmd
    assert "warning" in cmd
    assert "1000000loya" in cmd
    assert "True" in cmd


@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_cli_error(mock_subprocess_run):
    # Mock execution with non-zero return code
    mock_process = mock.MagicMock()
    mock_process.returncode = 1
    mock_process.stderr = b"Error message"
    mock_subprocess_run.return_value = mock_process

    result = propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    assert result is None


@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_transaction_code_error(mock_subprocess_run):
    # Mock execution with zero return code but non-zero tx code
    mock_process = mock.MagicMock()
    mock_process.returncode = 0
    mock_process.stdout = json.dumps({"code": 1, "raw_log": "Error in transaction"}).encode()
    mock_subprocess_run.return_value = mock_process

    result = propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    assert result is None


@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_exception_handling(mock_subprocess_run):
    # Mock execution with exception
    mock_subprocess_run.side_effect = Exception("Test exception")

    result = propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    assert result is None


@mock.patch("layer_values_monitor.dispute.logger")
@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_logs_success(mock_subprocess_run, mock_logger):
    # Test that successful execution is logged properly
    mock_process = mock.MagicMock()
    mock_process.returncode = 0
    mock_process.stdout = json.dumps({"code": 0, "txhash": "ABC123"}).encode()
    mock_subprocess_run.return_value = mock_process

    propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    mock_logger.info.assert_called_once_with("dispute msg executed successfully: ABC123")


@mock.patch("layer_values_monitor.dispute.logger")
@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_logs_cli_error(mock_subprocess_run, mock_logger):
    # Test that CLI errors are logged properly
    mock_process = mock.MagicMock()
    mock_process.returncode = 1
    mock_process.stderr = b"Error message"
    mock_subprocess_run.return_value = mock_process

    propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    mock_logger.error.assert_called_once()
    # Check that error contains stderr output
    assert "Error message" in str(mock_logger.error.call_args)


@mock.patch("layer_values_monitor.dispute.logger")
@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_logs_transaction_error(mock_subprocess_run, mock_logger):
    # Test that transaction errors are logged properly
    mock_process = mock.MagicMock()
    mock_process.returncode = 0
    mock_process.stdout = json.dumps({"code": 1, "raw_log": "Error in transaction"}).encode()
    mock_subprocess_run.return_value = mock_process

    propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    mock_logger.error.assert_called_once_with("failed to execute dispute msg: Error in transaction")


@mock.patch("layer_values_monitor.dispute.logger")
@mock.patch("layer_values_monitor.dispute.subprocess.run")
def test_logs_exception(mock_subprocess_run, mock_logger):
    # Test that exceptions are logged properly
    mock_subprocess_run.side_effect = Exception("Test exception")

    propose_msg(
        binary_path="layerd",
        reporter="reporter_address",
        query_id="query_id",
        meta_id="meta_id",
        dispute_category="warning",
        fee="1000000loya",
        key_name="key_name",
        chain_id="chain_id",
        rpc="http://localhost:26657",
        kb="test",
        kdir="/path/to/keyring",
        payfrom_bond="True",
    )

    mock_logger.error.assert_called_once_with("failed to execute dispute msg: Test exception")



import logging
from unittest.mock import MagicMock, patch

from layer_values_monitor.monitor import is_disputable

import pytest

magic_logger = MagicMock(spec=logging.Logger)


@pytest.mark.parametrize(
    "reported_value,trusted_value,alert_threshold,dispute_threshold,expected",
    [
        # Case: Percent difference exactly at alert threshold
        (110, 100, 0.1, 0.2, (True, False, 0.1)),
        # Case: Percent difference below alert threshold
        (105, 100, 0.1, 0.2, (False, False, 0.05)),
        # Case: Percent difference above alert threshold but below dispute threshold
        (115, 100, 0.1, 0.2, (True, False, 0.15)),
        # Case: Percent difference above both thresholds
        (125, 100, 0.1, 0.2, (True, True, 0.25)),
        # Case: Negative percent difference (reported < trusted)
        (80, 100, 0.1, 0.2, (True, True, 0.2)),
        # Case: dispute_threshold = 0 (special handling)
        (115, 100, 0.1, 0, (True, False, 0.15)),
    ],
)
def test_percentage_metric(reported_value, trusted_value, alert_threshold, dispute_threshold, expected):
    """Test percentage metric calculations and threshold checks."""
    with patch("layer_values_monitor.logger.logger.debug") as mock_logger:
        magic_logger.debug = mock_logger
        result = is_disputable("percentage", alert_threshold, dispute_threshold, reported_value, trusted_value, magic_logger)
        # logger.debug was called?
        mock_logger.assert_called_once()

        assert result[0] == expected[0]
        assert result[1] == expected[1]
        assert abs(result[2] - expected[2]) < 1e-10


@pytest.mark.parametrize(
    "reported_value,trusted_value,expected",
    [
        # Case: Values are equal (hex)
        ("0x123abc", "0x123abc", (False, False, 0.0)),
        # Case: Values are equal (different case)
        ("0x123ABC", "0x123abc", (False, False, 0.0)),
        # Case: Values are equal (one with prefix, one without)
        ("0x123abc", "123abc", (False, False, 0.0)),
        # Case: Values are not equal
        ("0x123abc", "0x456def", (True, True, 1.0)),
        (123, "123", (False, False, 0.0)),
        # Case: One value is None
        (None, "123abc", (True, True, 1.0)),
    ],
)
def test_equality_metric(reported_value, trusted_value, expected):
    """Test equality metric comparisons."""
    with patch("layer_values_monitor.logger.logger.info") as mock_logger:
        magic_logger.info = mock_logger
        result = is_disputable("equality", 0.1, 0.2, reported_value, trusted_value, magic_logger)

        mock_logger.assert_called_once()

        assert result == expected


@pytest.mark.parametrize(
    "reported_value,trusted_value,alert_threshold,dispute_threshold,expected",
    [
        # Case: Difference exactly at alert threshold
        (110, 100, 10, 20, (True, False, 10)),
        # Case: Difference below alert threshold
        (105, 100, 10, 20, (False, False, 5)),
        # Case: Difference above alert threshold but below dispute threshold
        (115, 100, 10, 20, (True, False, 15)),
        # Case: Difference above both thresholds
        (125, 100, 10, 20, (True, True, 25)),
        # Case: Negative difference (reported < trusted)
        (80, 100, 10, 20, (True, True, 20)),
        # Case: Reported and trusted are equal
        (100, 100, 10, 20, (False, False, 0)),
    ],
)
def test_range_metric(reported_value, trusted_value, alert_threshold, dispute_threshold, expected):
    """Test range metric calculations and threshold checks."""
    with patch("layer_values_monitor.logger") as mock_logger:
        result = is_disputable("range", alert_threshold, dispute_threshold, reported_value, trusted_value, mock_logger)

        # Range metric doesn't log anything specific
        mock_logger.debug.assert_not_called()
        mock_logger.info.assert_not_called()

        # Verify results
        assert result[0] == expected[0]
        assert result[1] == expected[1]
        assert result[2] == expected[2]


@pytest.mark.parametrize(
    "metric,case_variant",
    [
        ("percentage", "PERCENTAGE"),
        ("equality", "EqUaLiTy"),
        ("range", "Range"),
    ],
)
def test_case_insensitivity(metric, case_variant):
    """Test that metric string comparison is case-insensitive."""
    # Use simple values where results will be the same for all metrics
    result_lower = is_disputable(metric, 1, 1, 100, 100, magic_logger)
    result_variant = is_disputable(case_variant, 1, 1, 100, 100, magic_logger)

    assert result_lower == result_variant


def test_unknown_metric():
    """Test behavior with an unknown metric type."""
    result = is_disputable("unknown_metric", 0.1, 0.2, 100, 100, magic_logger)
    assert result[0] is None
